@@ -14,15 +14,13 @@
 from time import sleep
 
 from kuryr.lib import constants as kl_const
+from kuryr.lib import exceptions as kl_exc
 from kuryr.lib import segmentation_type_drivers as seg_driver
 from neutronclient.common import exceptions as n_exc
-from oslo_config import cfg as oslo_cfg
 from oslo_log import log as logging
 
 from kuryr_kubernetes import clients
-from kuryr_kubernetes import config
-from kuryr_kubernetes import constants as const
-from kuryr_kubernetes.controller.drivers import generic_vif
+from kuryr_kubernetes.controller.drivers import nested_vif
 from kuryr_kubernetes import exceptions as k_exc
 from kuryr_kubernetes import os_vif_util as ovu
 
@@ -33,8 +31,8 @@ DEFAULT_MAX_RETRY_COUNT = 3
 DEFAULT_RETRY_INTERVAL = 1
 
 
-class NestedVlanPodVIFDriver(generic_vif.GenericPodVIFDriver):
-    """Manages ports for nested-containers to provide VIFs."""
+class NestedVlanPodVIFDriver(nested_vif.NestedPodVIFDriver):
+    """Manages ports for nested-containers using VLANs to provide VIFs."""
 
     def request_vif(self, pod, project_id, subnets, security_groups):
         neutron = clients.get_neutron_client()
@@ -43,13 +41,65 @@ class NestedVlanPodVIFDriver(generic_vif.GenericPodVIFDriver):
 
         rq = self._get_port_request(pod, project_id, subnets, security_groups)
         port = neutron.create_port(rq).get('port')
-
         vlan_id = self._add_subport(neutron, trunk_id, port['id'])
 
-        vif_plugin = const.K8S_OS_VIF_NOOP_PLUGIN
-        vif = ovu.neutron_to_osvif_vif(vif_plugin, port, subnets)
-        vif.vlan_id = vlan_id
-        return vif
+        return ovu.neutron_to_osvif_vif_nested_vlan(port, subnets, vlan_id)
+
+    def request_vifs(self, pod, project_id, subnets, security_groups,
+                     num_ports):
+        """This method creates subports and returns a list with their vifs.
+
+        It creates up to num_ports subports and attaches them to the trunk
+        port.
+
+        If not enough vlan ids are available for all the subports to create,
+        it creates as much as available vlan ids.
+
+        Note the neutron trunk_add_subports is an atomic operation that will
+        either attach all or none of the subports. Therefore, if there is a
+        vlan id collision, all the created ports will be deleted and the
+        exception is raised.
+        """
+        neutron = clients.get_neutron_client()
+        parent_port = self._get_parent_port(neutron, pod)
+        trunk_id = self._get_trunk_id(parent_port)
+
+        port_rq, subports_info = self._create_subports_info(
+            pod, project_id, subnets, security_groups,
+            trunk_id, num_ports, unbound=True)
+
+        if not subports_info:
+            LOG.error("There are no vlan ids available to create subports")
+            return []
+
+        bulk_port_rq = {'ports': [port_rq for _ in range(len(subports_info))]}
+        try:
+            ports = neutron.create_port(bulk_port_rq).get('ports')
+        except n_exc.NeutronClientException as ex:
+            LOG.error("Error creating bulk ports: %s", bulk_port_rq)
+            raise ex
+        for index, port in enumerate(ports):
+            subports_info[index]['port_id'] = port['id']
+
+        try:
+            try:
+                neutron.trunk_add_subports(trunk_id,
+                                           {'sub_ports': subports_info})
+            except n_exc.Conflict as ex:
+                LOG.error("vlan ids already in use on trunk")
+                for port in ports:
+                    neutron.delete_port(port['id'])
+                raise ex
+        except n_exc.NeutronClientException as ex:
+            LOG.error("Error happened during subport addition to trunk")
+            raise ex
+
+        vifs = []
+        for index, port in enumerate(ports):
+            vlan_id = subports_info[index]['segmentation_id']
+            vif = ovu.neutron_to_osvif_vif_nested_vlan(port, subnets, vlan_id)
+            vifs.append(vif)
+        return vifs
 
     def release_vif(self, pod, vif):
         neutron = clients.get_neutron_client()
@@ -63,7 +113,8 @@ class NestedVlanPodVIFDriver(generic_vif.GenericPodVIFDriver):
             LOG.debug('Unable to release port %s as it no longer exists.',
                       vif.id)
 
-    def _get_port_request(self, pod, project_id, subnets, security_groups):
+    def _get_port_request(self, pod, project_id, subnets, security_groups,
+                          unbound=False):
         port_req_body = {'project_id': project_id,
                          'name': self._get_port_name(pod),
                          'network_id': self._get_network_id(subnets),
@@ -71,11 +122,36 @@ class NestedVlanPodVIFDriver(generic_vif.GenericPodVIFDriver):
                          'device_owner': kl_const.DEVICE_OWNER,
                          'admin_state_up': True}
 
+        if unbound:
+            port_req_body['name'] = 'available-port'
+
         if security_groups:
             port_req_body['security_groups'] = security_groups
 
         return {'port': port_req_body}
 
+    def _create_subports_info(self, pod, project_id, subnets,
+                              security_groups, trunk_id, num_ports,
+                              unbound=False):
+        subports_info = []
+
+        in_use_vlan_ids = self._get_in_use_vlan_ids_set(trunk_id)
+        port_rq = self._get_port_request(pod, project_id, subnets,
+                                         security_groups, unbound)['port']
+        for i in range(num_ports):
+            try:
+                vlan_id = seg_driver.allocate_segmentation_id(in_use_vlan_ids)
+            except kl_exc.SegmentationIdAllocationFailure:
+                LOG.warning("There is not enough vlan ids available to "
+                            "create a batch of %d subports.", num_ports)
+                break
+            in_use_vlan_ids.add(vlan_id)
+
+            subports_info.append({'segmentation_id': vlan_id,
+                                  'port_id': '',
+                                  'segmentation_type': 'vlan'})
+        return port_rq, subports_info
+
     def _get_trunk_id(self, port):
         try:
             return port['trunk_details']['trunk_id']
@@ -85,39 +161,6 @@ class NestedVlanPodVIFDriver(generic_vif.GenericPodVIFDriver):
                       "with a Neutron vlan trunk")
             raise k_exc.K8sNodeTrunkPortFailure
 
-    def _get_parent_port(self, neutron, pod):
-        node_subnet_id = config.CONF.neutron_defaults.worker_nodes_subnet
-        if not node_subnet_id:
-            raise oslo_cfg.RequiredOptError('worker_nodes_subnet',
-                    'neutron_defaults')
-
-        try:
-            # REVISIT(vikasc): Assumption is being made that hostIP is the IP
-            #		       of trunk interface on the node(vm).
-            node_fixed_ip = pod['status']['hostIP']
-        except KeyError:
-            if pod['status']['conditions'][0]['type'] != "Initialized":
-                LOG.debug("Pod condition type is not 'Initialized'")
-
-            LOG.error("Failed to get parent vm port ip")
-            raise
-
-        try:
-            fixed_ips = ['subnet_id=%s' % str(node_subnet_id),
-                         'ip_address=%s' % str(node_fixed_ip)]
-            ports = neutron.list_ports(fixed_ips=fixed_ips)
-        except n_exc.NeutronClientException as ex:
-            LOG.error("Parent vm port with fixed ips %s not found!",
-                      fixed_ips)
-            raise ex
-
-        if ports['ports']:
-            return ports['ports'][0]
-        else:
-            LOG.error("Neutron port for vm port with fixed ips %s"
-                      " not found!", fixed_ips)
-            raise k_exc.K8sNodeTrunkPortFailure
-
     def _add_subport(self, neutron, trunk_id, subport):
         """Adds subport port to Neutron trunk
 
@@ -165,7 +208,7 @@ class NestedVlanPodVIFDriver(generic_vif.GenericPodVIFDriver):
         subport_id = [{'port_id': subport_id}]
         try:
             neutron.trunk_remove_subports(trunk_id,
-                                       {'sub_ports': subport_id})
+                                          {'sub_ports': subport_id})
         except n_exc.NeutronClientException as ex:
             LOG.error(
                 "Error happened during subport removal from "
